#! /usr/bin/env python3

from dataclasses import dataclass
from math import floor
from typing import Optional

import numpy as np
from numpy.typing import ArrayLike, NDArray

import eclib.primeutils as pu
import eclib.randutils as ru


@dataclass(slots=True)
class PublicParameters:
    """Data class of cryptosystem parameters to be published.

    Attributes
    ----------
    p: int
        Safe prime
    q: int
        Sophie Germain prime
    g: int
        Generator of plaintext space

    Notes
    -----
    The plaintext space is a cyclib group G of order q, given by
    G = {g^i mod p | i = 1, ..., q}. The ciphertext space is G^2.

    """

    p: int
    q: int
    g: int

    def __init__(self, bit_length: Optional[int]):
        """Constructor for PublicParameters class.

        Generate a (bit_length)-bit Sophie Germain prime q, its associated safe prime
        p = 2q + 1, and a generator g of plaintext space such that g^q = 1 mod p. If
        bit_length is None, the parameters are initialized to 0.

        Parameters
        ----------
        bit_length : int, optional
            key length in bits.

        """
        if bit_length is None:
            self.p = self.q = self.g = 0

        else:
            self.q, self.p = pu.get_safe_prime(bit_length)
            self.g = _get_generator(self.q, self.p)


@dataclass(slots=True)
class SecretKey:
    """Data class of decryption key to be kept secret.

    Attributes
    ----------
    s: int
        Secret integer.

    """

    s: int

    def __init__(self, params: Optional[PublicParameters]):
        """Constructor for SecretKey class.

        Generate a random integer s in {1, ... q - 1} used for decryption, where q is a
        cryptosystem parameter. If params is None, s is initialized to 0.

        Parameters
        ----------
        params : PublicParameters, optional
            Cryptosystem parameters.

        """
        if params is None:
            self.s = 0

        else:
            self.s = ru.get_rand(1, params.q)


@dataclass(slots=True)
class PublicKey:
    """Data class of encryption key to be published.

    Attributes
    ----------
    h: int
        Public integer

    """

    h: int

    def __init__(self, params: Optional[PublicParameters], sk: Optional[SecretKey]):
        """Constructor for PublicKey class.

        Compute an integer h = g^s mod p used for encryption, where s is a secret
        integer used for decryption, and p is a cryptosystem parameter. If params and
        s are None, h is initialized to 0.

        Parameters
        ----------
        params : PublicParameters, optional
            Cryptosystem parameters.
        sk : SecretKey, optional
            Decryption key.

        """
        if params is None and sk is None:
            self.h = 0

        elif params is not None and sk is not None:
            self.h = pow(params.g, sk.s, params.p)


def keygen(bit_length: int) -> tuple[PublicParameters, PublicKey, SecretKey]:
    """Key generation function.

    Provide cryptosystem parameters, an encryption key, and a decryption key.

    Parameters
    ----------
    bit_length : int
        Key length in bits.

    Returns
    -------
    tuple[PublicParameters, PublicKey, SecretKey]
        Cryptosystem parameters, encryption key, and decryption key.

    """
    params = PublicParameters(bit_length)

    sk = SecretKey(params)

    pk = PublicKey(params, sk)

    return params, pk, sk


def encrypt(
    params: PublicParameters, pk: PublicKey, m: ArrayLike
) -> NDArray[np.object_]:
    """Encryption function

    Provide an encrypted message of a message m. If m is a vector or matrix, encrypt
    each element of m separately and return a vector or matrix encrypted message.

    Parameters
    ----------
    params : PublicParameters
        Cryptosystem parameters.
    pk : PublicKey
        Encryption key.
    m : ArrayLike
        Message in a plaintext space.

    Returns
    -------
    NDArray[np.object_]
        Encrypted message in a ciphertext space.

    Raises
    ------
    ValueError
        If the message is not scalar, vector, or matrix.

    """
    m = np.asarray(m, dtype=object)

    match m.ndim:
        case 0:
            return _encrypt(params, pk, m.item())

        case 1:
            return np.array(
                [_encrypt(params, pk, m[i]) for i in range(m.shape[0])],
                dtype=object,
            )

        case 2:
            return np.array(
                [
                    [_encrypt(params, pk, m[i][j]) for j in range(m.shape[1])]
                    for i in range(m.shape[0])
                ],
                dtype=object,
            )

        case _:
            raise ValueError


def decrypt(
    params: PublicParameters, sk: SecretKey, c: NDArray[np.object_]
) -> ArrayLike:
    """Decryption function

    Provide the recovered message of an encrypted message c. If c is a vector or
    matrix, decrypt each element of c separately and return a vector or matrix
    recovered message.

    Parameters
    ----------
    params : PublicParameters
        Cryptosystem parameters.
    sk : SecretKey
        Decryption key.
    c : NDArray[np.object_]
        Encrypted message in a ciphertext space.

    Returns
    -------
    ArrayLike
        Message in a plaintext space.

    Raises
    ------
    ValueError
        If the encrypted message is not scalar, vector, or matrix.

    """
    c = np.asarray(c, dtype=object)

    match c.ndim - 1:
        case 0:
            return _decrypt(params, sk, c)

        case 1:
            return np.array(
                [_decrypt(params, sk, c[i]) for i in range(c.shape[0])],
                dtype=object,
            )

        case 2:
            return np.array(
                [
                    [_decrypt(params, sk, c[i][j]) for j in range(c.shape[1])]
                    for i in range(c.shape[0])
                ],
                dtype=object,
            )

        case _:
            raise ValueError


def mult(
    params: PublicParameters, c1: NDArray[np.object_], c2: NDArray[np.object_]
) -> NDArray[np.object_]:
    """Homomorphic multiplication function

    Provide an encrypted message of the Hadamard product of m1 and m2 from ciphertexts
    c1 and c2 corresponding to the messages m1 and m2. This function supports the
    Hadamard product of the following forms: scalar-scalar, scalar-vector,
    scalar-matrix, vector-vector, matrix-vector, and matrix-matrix.

    Parameters
    ----------
    params : PublicParameters
        Cryptosystem parameters.
    c1 : NDArray[np.object_]
        Encrypted message in a ciphertext space.
    c2 : NDArray[np.object_]
        Encrypted message in a ciphertext space.

    Returns
    -------
    NDArray[np.object_]
        Encrypted message in a ciphertext space.

    Raises
    ------
    ValueError
        If the size of the encrypted messages are not appropriate.

    """
    c1 = np.asarray(c1, dtype=object)
    c2 = np.asarray(c2, dtype=object)

    match c1.ndim - 1:
        case 0 if c2.ndim - 1 == 0:
            return _mult(params, c1, c2)

        case 0 if c2.ndim - 1 == 1:
            return np.array(
                [_mult(params, c1, c2[i]) for i in range(c2.shape[0])],
                dtype=object,
            )

        case 0 if c2.ndim - 1 == 2:
            return np.array(
                [
                    [_mult(params, c1, c2[i][j]) for j in range(c2.shape[1])]
                    for i in range(c2.shape[0])
                ],
                dtype=object,
            )

        case 1 if c1.shape == c2.shape:
            return np.array(
                [_mult(params, c1[i], c2[i]) for i in range(c1.shape[0])],
                dtype=object,
            )

        case 2 if c2.ndim - 1 == 1 and c1.shape[1] == c2.shape[0]:
            return np.array(
                [
                    [_mult(params, c1[i][j], c2[j]) for j in range(c1.shape[1])]
                    for i in range(c1.shape[0])
                ],
                dtype=object,
            )

        case 2 if c1.shape == c2.shape:
            return np.array(
                [
                    [_mult(params, c1[i][j], c2[i][j]) for j in range(c1.shape[1])]
                    for i in range(c1.shape[0])
                ],
                dtype=object,
            )

        case _:
            raise ValueError


def encode(params: PublicParameters, x: ArrayLike, delta: float) -> ArrayLike:
    """Encoding function

    Provide a message that can be encrypted from a real-valued data x. The message is
    the nearest element in a plaintext space of delta * x. If x is a vector or matrix,
    encode each element of x separately and return a vector or matrix message.

    Parameters
    ----------
    params : PublicParameters
        Cryptosystem parameters.
    x : ArrayLike
        Real-valued data.
    delta : float
        Scaling parameter.

    Returns
    -------
    ArrayLike
        Message in a plaintext space.

    """
    f = np.frompyfunc(_encode, 3, 1)
    return f(params, x, delta)


def decode(params: PublicParameters, m: ArrayLike, delta: float) -> ArrayLike:
    """Decoding function

    Parameters
    ----------
    params : PublicParameters
        Cryptosystem parameters.
    m : ArrayLike
        Message in a plaintext space.
    delta : float
        Scaling parameter.

    Returns
    -------
    ArrayLike
        Real-valued data.

    """
    f = np.frompyfunc(_decode, 3, 1)
    return f(params, m, delta)


def enc(
    params: PublicParameters, pk: PublicKey, x: ArrayLike, delta: float
) -> NDArray[np.object_]:
    """Encoding and Encryption function

    Provide an encrypted message of a message corresponding to a real-valued data x.
    If x is a vector or matrix, encode and encrypt each element of x separately and
    return a vector or matrix encrypted message.

    Parameters
    ----------
    params : PublicParameters
        Cryptosystem parameters.
    pk : PublicKey
        Encryption key.
    x : ArrayLike
        Real-valued data.
    delta : float
        Scaling parameter.

    Returns
    -------
    NDArray[np.object_]
        Encrypted message in a ciphertext space.

    See Also
    --------
    encode
    encrypt

    """
    return encrypt(params, pk, encode(params, x, delta))


def dec(
    params: PublicParameters, sk: SecretKey, c: NDArray[np.object_], delta: float
) -> ArrayLike:
    """Decryption and decoding function

    Provide the real-valued data corresponding to the recovered message of an encrypted
    message c. If c is a vector or matrix, decrypt and decode each element of c
    separately and return a vector or matrix recovered message.

    Parameters
    ----------
    params : PublicParameters
        Cryptosystem parameters.
    sk : SecretKey
        Decryption key.
    c : NDArray[np.object_]
        Encrypted message in a ciphertext space.
    delta : float
        Scaling parameter.

    Returns
    -------
    ArrayLike
        Message in a plaintext space.

    See Also
    --------
    decode
    decrypt
    """
    return decode(params, decrypt(params, sk, c), delta)


def dec_add(
    params: PublicParameters, sk: SecretKey, c: NDArray[np.object_], delta: float
) -> ArrayLike:
    c = np.asarray(c, dtype=object)

    match c.ndim - 1:
        case 0:
            return dec(params, sk, c, delta)

        case 1:
            return np.sum(dec(params, sk, c, delta), axis=0)

        case 2:
            return np.sum(dec(params, sk, c, delta), axis=1)

        case _:
            raise ValueError


def _encrypt(params: PublicParameters, pk: PublicKey, m: int) -> NDArray[np.object_]:
    r = ru.get_rand(1, params.q)

    return np.array(
        [pow(params.g, r, params.p), (m * pow(pk.h, r, params.p)) % params.p],
        dtype=object,
    )


def _decrypt(params: PublicParameters, sk: SecretKey, c: NDArray[np.object_]) -> int:
    return (pow(c[0], -sk.s, params.p) * c[1]) % params.p


def _mult(
    params: PublicParameters, c1: NDArray[np.object_], c2: NDArray[np.object_]
) -> NDArray[np.object_]:
    return np.array(
        [(c1[0] * c2[0]) % params.p, (c1[1] * c2[1]) % params.p], dtype=object
    )


def _encode(params: PublicParameters, x: float, delta: float) -> int:
    m = floor(x / delta + 0.5)
    first_decimal_place = (x / delta * 10) % 10

    if m < 0:
        if m < -params.q:
            raise ValueError("Underflow")
        else:
            m += params.p
    elif m > params.q:
        raise ValueError("Overflow")

    if x / delta == int(x / delta) or first_decimal_place >= 5:
        for i in range(params.q):
            if m - i > 0 and _is_element(m - i, params.q, params.p):
                return m - i
            elif m + i < params.p and _is_element(m + i, params.q, params.p):
                return m + i
    else:
        for i in range(params.q):
            if m + i < params.p and _is_element(m + i, params.q, params.p):
                return m + i
            elif m - i > 0 and _is_element(m - i, params.q, params.p):
                return m - i

    raise ValueError


def _decode(params: PublicParameters, m: int, delta: float) -> float:
    if m > params.q:
        return (m - params.p) * delta

    else:
        return m * delta


def _is_generator(g: int, q: int, p: int) -> bool:
    if g <= 1 or g >= p:
        return False
    elif pow(g, q, p) == 1:
        return True
    else:
        return False


def _is_element(m: int, q: int, p: int) -> bool:
    if m <= 0 or m >= p:
        return False
    elif pow(m, q, p) == 1:
        return True
    else:
        return False


def _get_generator(q: int, p: int) -> int:
    g = 2
    while _is_generator(g, q, p) is False:
        g += 1

    return g
